"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = createGenerator;

var _cssInJsUtils = require("css-in-js-utils");

var _color = _interopRequireDefault(require("color"));

var _adapters = _interopRequireDefault(require("./adapters"));

var _hash = _interopRequireDefault(require("./hash"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var validPseudoClasses = ['link', 'hover', 'focus', 'active', 'visited', 'disabled', 'focusWithin', 'firstChild', 'lastChild'];
var validMediaQueries = ['viewportWidth', 'viewportHeight'];

function createGenerator() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$adapter = _ref.adapter,
      adapter = _ref$adapter === void 0 ? 'fela' : _ref$adapter,
      _ref$devMode = _ref.devMode,
      devMode = _ref$devMode === void 0 ? false : _ref$devMode,
      _ref$rootNode = _ref.rootNode,
      rootNode = _ref$rootNode === void 0 ? 'body' : _ref$rootNode,
      _ref$dynamicImport = _ref.dynamicImport,
      dynamicImport = _ref$dynamicImport === void 0 ? false : _ref$dynamicImport;

  var usedAdapter = _adapters.default.find(function (adapt) {
    return adapt.name === adapter;
  });

  var config = {
    devMode: devMode,
    rootNode: rootNode,
    dynamicImport: dynamicImport
  };
  return function generate(ast, fileName) {
    var css = generateCSS(ast, config);
    var js = generateJS(ast, config, usedAdapter);
    var root = generateRoot(ast, config);
    return _objectSpread(_defineProperty({}, fileName + '.js', root), css, js);
  };
}

function getModuleName(module, devMode) {
  var hashedBody = '_' + (0, _hash.default)(JSON.stringify(module.body));

  if (devMode) {
    return module.name + hashedBody;
  }

  return hashedBody;
}

function generateRoot(ast) {
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  var imports = styles.map(function (module) {
    return 'import { ' + module.name + " } from './" + module.name + ".elo.js'";
  }).join('\n');
  return imports + '\n\n' + 'export {\n  ' + styles.map(function (module) {
    return module.name;
  }).join(',\n  ') + '\n}';
}

function generateCSSValue(value, property) {
  var unit = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;

  if (value.type === 'Integer') {
    return (value.negative ? '-' : '') + value.value + (unit && !(0, _cssInJsUtils.isUnitlessProperty)(property) ? 'px' : '');
  }

  if (value.type === 'RawValue' || value.type === 'String') {
    return value.value;
  }

  if (value.type === 'Percentage') {
    if (property === 'opacity') {
      return value.value / 100;
    } else {
      return value.value + '%';
    }
  }

  if (value.type === 'Color') {
    var format = value.format,
        red = value.red,
        blue = value.blue,
        green = value.green,
        alpha = value.alpha;

    var colorValue = _color.default.rgb(red, green, blue, alpha);

    if (format === 'hex') {
      return colorValue.hex();
    }

    if (format === 'keyword') {
      // TODO: check APIs
      return colorValue.keyword();
    }

    return colorValue[format]().string();
  }

  if (value.type === 'Float') {
    return (value.negative ? '-' : '') + value.integer + '.' + value.fractional + (unit ? 'px' : '');
  }

  if (value.type === 'Identifier') {
    return (0, _cssInJsUtils.hyphenateProperty)(value.value);
  }

  return value.value;
}

function generateCSS(ast, _ref2) {
  var devMode = _ref2.devMode;
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  var variants = ast.body.filter(function (node) {
    return node.type === 'Variant';
  });
  return styles.reduce(function (files, module) {
    var classes = generateClasses(module.body, variants);
    files[module.name + '.elo.css'] = classes.filter(function (selector) {
      return selector.declarations.length > 0;
    }).map(function (selector) {
      var css = stringifyCSS(selector.declarations, getModuleName(module, devMode) + selector.modifier + selector.pseudo);

      if (selector.media) {
        return '@media ' + selector.media + ' {\n' + css + '\n}';
      }

      return css;
    }).join('\n\n');
    return files;
  }, {});
}

function generateClasses(nodes, variants) {
  var classes = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
  var modifier = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : '';
  var pseudo = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : '';
  var media = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : '';
  var base = nodes.filter(function (node) {
    return node.type === 'Declaration';
  });
  var nesting = nodes.filter(function (node) {
    return node.type !== 'Declaration';
  });
  classes.push({
    media: media,
    pseudo: pseudo,
    modifier: modifier,
    declarations: getStaticDeclarations(base)
  });
  nesting.forEach(function (nest) {
    if (nest.property.type === 'Identifier') {
      var variant = variants.find(function (variant) {
        return variant.name === nest.property.value;
      });

      if (variant) {
        if (nest.value.type === 'Identifier') {
          var variation = variant.body.find(function (variant) {
            return variant.value === nest.value.value;
          });

          if (variation) {
            generateClasses(nest.body, variants, classes, modifier + '__' + variant.name + '-' + variation.value, pseudo, media);
          }
        }
      } else {// TODO: throw
      }
    }

    if (nest.property.type === 'Variable' && nest.property.environment) {
      if (nest.boolean && validPseudoClasses.indexOf(nest.property.value) !== -1) {
        generateClasses(nest.body, variants, classes, modifier, pseudo + ':' + (0, _cssInJsUtils.hyphenateProperty)(nest.property.value), media);
      }

      if (validMediaQueries.indexOf(nest.property.value) !== -1) {
        generateClasses(nest.body, variants, classes, modifier, pseudo, getMediaQuery(nest.value.value, nest.property.value, nest.operator));
      }
    }
  });
  return classes;
}

function getStaticDeclarations(declarations) {
  return declarations.filter(function (decl) {
    return !decl.dynamic;
  }).map(function (declaration) {
    return {
      property: declaration.property,
      value: generateCSSValue(declaration.value, declaration.property)
    };
  });
}

function stringifyCSS(declarations, name) {
  return '.' + name + ' {\n  ' + declarations.map(function (decl) {
    return (0, _cssInJsUtils.hyphenateProperty)(decl.property) + ': ' + decl.value;
  }).join(';\n  ') + '\n' + '}';
}

function generateJS(ast, _ref3, adapter) {
  var devMode = _ref3.devMode,
      dynamicImport = _ref3.dynamicImport;
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  var variants = ast.body.filter(function (node) {
    return node.type === 'Variant';
  });
  return styles.reduce(function (files, module) {
    var style = generateStyle(module.body);
    var classNameMap = generateClassNameMap(module.body, variants);
    files[module.name + '.elo.js'] = adapter.stringify({
      style: style,
      moduleName: module.name,
      dynamicImport: dynamicImport,
      classNameMap: classNameMap,
      className: '_elo_' + module.format + ' ' + getModuleName(module, devMode),
      variants: variants.reduce(function (flatVariants, variant) {
        flatVariants[variant.name] = variant.body.map(function (variation) {
          return variation.value;
        });
        return flatVariants;
      }, {})
    });
    return files;
  }, {});
}

function generateClassNameMap(nodes, variants) {
  var classes = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  var variations = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  var modifier = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : '';
  var base = nodes.filter(function (node) {
    return node.type === 'Declaration';
  });
  var nesting = nodes.filter(function (node) {
    return node.type !== 'Declaration';
  });
  classes[modifier] = variations;
  nesting.forEach(function (nest) {
    if (nest.property.type === 'Identifier') {
      var variant = variants.find(function (variant) {
        return variant.name === nest.property.value;
      });

      if (variant) {
        if (nest.value.type === 'Identifier') {
          var variation = variant.body.find(function (variant) {
            return variant.value === nest.value.value;
          });

          if (variation) {
            generateClassNameMap(nest.body, variants, classes, _objectSpread({}, variations, _defineProperty({}, variant.name, variation.value)), modifier + '__' + variant.name + '-' + variation.value);
          }
        }
      } else {// TODO: throw
      }
    }
  });
  return classes;
}

function generateStyle(nodes) {
  var base = nodes.filter(function (node) {
    return node.type === 'Declaration';
  });
  var nestings = nodes.filter(function (node) {
    return node.type !== 'Declaration';
  });
  var declarations = base.filter(function (decl) {
    return decl.dynamic;
  }).map(function (declaration) {
    return {
      property: declaration.property,
      value: declaration.value.value
    };
  });
  var nests = nestings.map(function (nest) {
    if (nest.property.type === 'Variable' && nest.property.environment) {
      if (nest.boolean && validPseudoClasses.indexOf(nest.property.value) !== -1) {
        return {
          property: ':' + nest.property.value,
          value: generateStyle(nest.body)
        };
      }

      if (validMediaQueries.indexOf(nest.property.value) !== -1) {
        return {
          property: getMediaQuery(nest.value.value, nest.property.value, nest.operator),
          value: generateStyle(nest.body)
        };
      }
    }
  }).filter(function (nesting) {
    return nesting && nesting.value.length > 0;
  });
  return [].concat(_toConsumableArray(declarations), _toConsumableArray(nests));
}

function getMediaQuery(value, property, operator) {
  var dimension = property.indexOf('Height') !== -1 ? 'height' : 'width';

  if (operator === '=') {
    return '(min-' + dimension + ': ' + value + 'px) and (max-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '>') {
    return '(min-' + dimension + ': ' + (value + 1) + 'px)';
  }

  if (operator === '>=') {
    return '(min-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '<=') {
    return '(max-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '<') {
    return '(max-' + dimension + ': ' + (value - 1) + 'px)';
  }
}